@@ -13,7 +13,13 @@ import pytest
 import requests
 import responses
 from requests.exceptions import ConnectionError, HTTPError
-from responses import BaseResponse, Response, matchers
+from responses import (
+    BaseResponse,
+    Response,
+    PassthroughResponse,
+    matchers,
+    CallbackResponse,
+)
 
 
 try:
@@ -94,6 +100,9 @@ def test_response_with_instance():
         assert len(responses.calls) == 2
         assert responses.calls[1].request.url == "http://example.com/?foo=bar"
 
+    run()
+    assert_reset()
+
 
 @pytest.mark.parametrize(
     "original,replacement",
@@ -334,6 +343,27 @@ def test_match_querystring():
     assert_reset()
 
 
+def test_query_string_matcher():
+    @responses.activate
+    def run():
+        url = "http://example.com?test=1&foo=bar"
+        responses.add(
+            responses.GET,
+            url,
+            body=b"test",
+            match=[matchers.query_string_matcher("test=1&foo=bar")],
+        )
+        resp = requests.get("http://example.com?test=1&foo=bar")
+        assert_response(resp, "test")
+        resp = requests.get("http://example.com?foo=bar&test=1")
+        assert_response(resp, "test")
+        resp = requests.get("http://example.com/?foo=bar&test=1")
+        assert_response(resp, "test")
+
+    run()
+    assert_reset()
+
+
 def test_match_empty_querystring():
     @responses.activate
     def run():
@@ -551,6 +581,11 @@ def test_callback():
     assert_reset()
 
 
+def test_callback_deprecated_argument():
+    with pytest.deprecated_call():
+        CallbackResponse(responses.GET, "url", lambda x: x, stream=False)
+
+
 def test_callback_exception_result():
     result = Exception()
     url = "http://example.com/"
@@ -899,7 +934,7 @@ def test_response_filebody():
         with responses.RequestsMock() as m:
             with open(current_file, "r") as out:
                 m.add(responses.GET, "http://example.com", body=out.read(), stream=True)
-                resp = requests.get("http://example.com")
+                resp = requests.get("http://example.com", stream=True)
             with open(current_file, "r") as out:
                 assert resp.text == out.read()
 
@@ -907,6 +942,18 @@ def test_response_filebody():
     assert_reset()
 
 
+def test_use_stream_twice_to_double_raw_io():
+    @responses.activate
+    def run():
+        url = "http://example.com"
+        responses.add(responses.GET, url, body=b"42", stream=True)
+        resp = requests.get(url, stream=True)
+        assert resp.raw.read() == b"42"
+
+    run()
+    assert_reset()
+
+
 def test_assert_all_requests_are_fired():
     def request_callback(request):
         raise BaseException()
@@ -1245,6 +1292,90 @@ def test_multiple_methods():
     assert_reset()
 
 
+def test_passthrough_flag(httpserver):
+    httpserver.serve_content("OK", headers={"Content-Type": "text/plain"})
+    response = Response(responses.GET, httpserver.url, body="MOCK")
+
+    @responses.activate
+    def run_passthrough():
+        responses.add(response)
+        resp = requests.get(httpserver.url)
+        assert_response(resp, "OK")
+
+    @responses.activate
+    def run_mocked():
+        responses.add(response)
+        resp = requests.get(httpserver.url)
+        assert_response(resp, "MOCK")
+
+    run_mocked()
+    assert_reset()
+
+    response.passthrough = True
+    run_passthrough()
+    assert_reset()
+
+
+def test_passthrough_response(httpserver):
+    httpserver.serve_content("OK", headers={"Content-Type": "text/plain"})
+
+    @responses.activate
+    def run():
+        responses.add(PassthroughResponse(responses.GET, httpserver.url))
+        responses.add(responses.GET, "{}/one".format(httpserver.url), body="one")
+        responses.add(responses.GET, "http://example.com/two", body="two")
+
+        resp = requests.get("http://example.com/two")
+        assert_response(resp, "two")
+        resp = requests.get("{}/one".format(httpserver.url))
+        assert_response(resp, "one")
+        resp = requests.get(httpserver.url)
+        assert_response(resp, "OK")
+
+        assert len(responses.calls) == 3
+        responses.assert_call_count(httpserver.url, 1)
+
+    run()
+    assert_reset()
+
+
+def test_passthrough_response_stream(httpserver):
+    httpserver.serve_content("OK", headers={"Content-Type": "text/plain"})
+
+    @responses.activate
+    def run():
+        responses.add(PassthroughResponse(responses.GET, httpserver.url))
+        content_1 = requests.get(httpserver.url).content
+        with requests.get(httpserver.url, stream=True) as resp:
+            content_2 = resp.raw.read()
+        assert content_1 == content_2
+
+    run()
+    assert_reset()
+
+
+def test_passthru_prefixes(httpserver):
+    httpserver.serve_content("OK", headers={"Content-Type": "text/plain"})
+
+    @responses.activate
+    def run_constructor_argument():
+        with responses.RequestsMock(passthru_prefixes=(httpserver.url,)):
+            resp = requests.get(httpserver.url)
+            assert_response(resp, "OK")
+
+    @responses.activate
+    def run_property_setter():
+        with responses.RequestsMock() as m:
+            m.passthru_prefixes = tuple([httpserver.url])
+            resp = requests.get(httpserver.url)
+            assert_response(resp, "OK")
+
+    run_constructor_argument()
+    assert_reset()
+    run_property_setter()
+    assert_reset()
+
+
 def test_passthru(httpserver):
     httpserver.serve_content("OK", headers={"Content-Type": "text/plain"})
 
@@ -1457,37 +1588,81 @@ def test_request_matches_post_params():
         )
         assert_response(resp, "one")
 
-    for depr in [True, False]:
-        run(deprecated=depr)
+    with pytest.deprecated_call():
+        run(deprecated=True)
         assert_reset()
 
+    run(deprecated=False)
+    assert_reset()
+
 
 def test_request_matches_empty_body():
-    @responses.activate
     def run():
-        responses.add(
-            method=responses.POST,
-            url="http://example.com/",
-            body="one",
-            match=[matchers.json_params_matcher(None)],
-        )
+        with responses.RequestsMock(assert_all_requests_are_fired=True) as rsps:
+            # test that both json and urlencoded body are empty in matcher and in request
+            rsps.add(
+                method=responses.POST,
+                url="http://example.com/",
+                body="one",
+                match=[matchers.json_params_matcher(None)],
+            )
 
-        responses.add(
-            method=responses.POST,
-            url="http://example.com/",
-            body="two",
-            match=[matchers.urlencoded_params_matcher(None)],
-        )
+            rsps.add(
+                method=responses.POST,
+                url="http://example.com/",
+                body="two",
+                match=[matchers.urlencoded_params_matcher(None)],
+            )
 
-        resp = requests.request("POST", "http://example.com/")
-        assert_response(resp, "one")
+            resp = requests.request("POST", "http://example.com/")
+            assert_response(resp, "one")
 
-        resp = requests.request(
-            "POST",
-            "http://example.com/",
-            headers={"Content-Type": "x-www-form-urlencoded"},
-        )
-        assert_response(resp, "two")
+            resp = requests.request(
+                "POST",
+                "http://example.com/",
+                headers={"Content-Type": "x-www-form-urlencoded"},
+            )
+            assert_response(resp, "two")
+
+        with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
+            # test exception raise if matcher body is None but request data is not None
+            rsps.add(
+                method=responses.POST,
+                url="http://example.com/",
+                body="one",
+                match=[matchers.json_params_matcher(None)],
+            )
+
+            with pytest.raises(ConnectionError) as excinfo:
+                resp = requests.request(
+                    "POST",
+                    "http://example.com/",
+                    json={"my": "data"},
+                    headers={"Content-Type": "application/json"},
+                )
+
+            msg = str(excinfo.value)
+            assert "request.body doesn't match: {my: data} doesn't match {}" in msg
+
+        with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
+            rsps.add(
+                method=responses.POST,
+                url="http://example.com/",
+                body="two",
+                match=[matchers.urlencoded_params_matcher(None)],
+            )
+            with pytest.raises(ConnectionError) as excinfo:
+                resp = requests.request(
+                    "POST",
+                    "http://example.com/",
+                    headers={"Content-Type": "x-www-form-urlencoded"},
+                    data={"page": "second", "type": "urlencoded"},
+                )
+            msg = str(excinfo.value)
+            assert (
+                "request.body doesn't match: {page: second, type: urlencoded} doesn't match {}"
+                in msg
+            )
 
     run()
     assert_reset()
@@ -1526,76 +1701,349 @@ def test_request_matches_params():
     assert_reset()
 
 
-def test_fail_request_error():
+def test_multipart_matcher():
     @responses.activate
     def run():
-        responses.add("POST", "http://example1.com")
-        responses.add("GET", "http://example.com")
+        req_data = {"some": "other", "data": "fields"}
+        req_files = {"file_name": b"Old World!"}
         responses.add(
-            "POST",
-            "http://example.com",
-            match=[matchers.urlencoded_params_matcher({"foo": "bar"})],
-        )
-        responses.add(
-            "POST",
-            "http://example.com",
-            match=[matchers.json_params_matcher({"fail": "json"})],
+            responses.POST,
+            url="http://httpbin.org/post",
+            match=[matchers.multipart_matcher(req_files, data=req_data)],
         )
+        resp = requests.post("http://httpbin.org/post", data=req_data, files=req_files)
+        assert resp.status_code == 200
 
-        with pytest.raises(ConnectionError) as excinfo:
-            requests.post("http://example.com", data={"id": "bad"})
-
-        msg = str(excinfo.value)
-        assert "- POST http://example1.com/ URL does not match" in msg
-        assert "- GET http://example.com/ Method does not match" in msg
-
-        if six.PY3:
-            assert "Parameters do not match. id=bad doesn't match {'foo': 'bar'}" in msg
-        else:
-            assert (
-                "Parameters do not match. id=bad doesn't match {u'foo': u'bar'}" in msg
+        with pytest.raises(TypeError):
+            responses.add(
+                responses.POST,
+                url="http://httpbin.org/post",
+                match=[matchers.multipart_matcher(files={})],
             )
 
-        assert (
-            "Parameters do not match. JSONDecodeError: Cannot parse request.body" in msg
-        )
+    run()
+    assert_reset()
 
-        # test query parameters
+
+def test_fail_request_error():
+    """
+    Validate that exception is raised if request URL/Method/kwargs don't match
+    :return:
+    """
+
+    def run():
+        with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
+            rsps.add("POST", "http://example1.com")
+            rsps.add("GET", "http://example.com")
+
+            with pytest.raises(ConnectionError) as excinfo:
+                requests.post("http://example.com", data={"id": "bad"})
+
+            msg = str(excinfo.value)
+            assert "- POST http://example1.com/ URL does not match" in msg
+            assert "- GET http://example.com/ Method does not match" in msg
+
+    run()
+    assert_reset()
+
+
+def test_fail_matchers_error():
+    """
+    Validate that Exception is raised if request does not match responses.matchers
+        validate matchers.urlencoded_params_matcher
+        validate matchers.json_params_matcher
+        validate matchers.query_param_matcher
+        validate matchers.request_kwargs_matcher
+    :return: None
+    """
+
+    def run():
+        with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
+            rsps.add(
+                "POST",
+                "http://example.com",
+                match=[matchers.urlencoded_params_matcher({"foo": "bar"})],
+            )
+            rsps.add(
+                "POST",
+                "http://example.com",
+                match=[matchers.json_params_matcher({"fail": "json"})],
+            )
+
+            with pytest.raises(ConnectionError) as excinfo:
+                requests.post("http://example.com", data={"id": "bad"})
+
+            msg = str(excinfo.value)
+            assert (
+                "request.body doesn't match: {id: bad} doesn't match {foo: bar}" in msg
+            )
+
+            assert (
+                "request.body doesn't match: JSONDecodeError: Cannot parse request.body"
+                in msg
+            )
+
+        with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
+            rsps.add(
+                "GET",
+                "http://111.com",
+                match=[matchers.query_param_matcher({"my": "params"})],
+            )
+
+            rsps.add(
+                method=responses.GET,
+                url="http://111.com/",
+                body="two",
+                match=[matchers.json_params_matcher({"page": "one"})],
+            )
+
+            with pytest.raises(ConnectionError) as excinfo:
+                requests.get(
+                    "http://111.com", params={"id": "bad"}, json={"page": "two"}
+                )
+
+            msg = str(excinfo.value)
+            assert (
+                "Parameters do not match. {id: bad} doesn't match {my: params}" in msg
+            )
+            assert (
+                "request.body doesn't match: {page: two} doesn't match {page: one}"
+                in msg
+            )
+
+        with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
+            req_kwargs = {
+                "stream": True,
+                "verify": False,
+            }
+            rsps.add(
+                "GET",
+                "http://111.com",
+                match=[matchers.request_kwargs_matcher(req_kwargs)],
+            )
+
+            with pytest.raises(ConnectionError) as excinfo:
+                requests.get("http://111.com", stream=True)
+
+            msg = str(excinfo.value)
+            assert (
+                "Arguments don't match: "
+                "{stream: True, verify: True} doesn't match {stream: True, verify: False}"
+            ) in msg
+
+    run()
+    assert_reset()
+
+
+def test_fail_multipart_matcher():
+    """
+    Validate that Exception is raised if request does not match responses.matchers
+        validate matchers.multipart_matcher
+    :return: None
+    """
+
+    def run():
+        # different file contents
+        with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
+            req_data = {"some": "other", "data": "fields"}
+            req_files = {"file_name": b"Old World!"}
+            rsps.add(
+                responses.POST,
+                url="http://httpbin.org/post",
+                match=[matchers.multipart_matcher(req_files, data=req_data)],
+            )
+
+            with pytest.raises(ConnectionError) as excinfo:
+                requests.post(
+                    "http://httpbin.org/post",
+                    data=req_data,
+                    files={"file_name": b"New World!"},
+                )
+
+            msg = str(excinfo.value)
+            assert "multipart/form-data doesn't match. Request body differs." in msg
+            assert (
+                '\r\nContent-Disposition: form-data; name="file_name"; '
+                'filename="file_name"\r\n\r\nOld World!\r\n'
+            ) in msg
+            assert (
+                '\r\nContent-Disposition: form-data; name="file_name"; '
+                'filename="file_name"\r\n\r\nNew World!\r\n'
+            ) in msg
+
+        # x-www-form-urlencoded request
+        with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
+            req_data = {"some": "other", "data": "fields"}
+            req_files = {"file_name": b"Old World!"}
+            rsps.add(
+                responses.POST,
+                url="http://httpbin.org/post",
+                match=[matchers.multipart_matcher(req_files, data=req_data)],
+            )
+
+            with pytest.raises(ConnectionError) as excinfo:
+                requests.post("http://httpbin.org/post", data=req_data)
+
+            msg = str(excinfo.value)
+            assert (
+                "multipart/form-data doesn't match. Request headers['Content-Type'] is different."
+                in msg
+            )
+            assert (
+                "application/x-www-form-urlencoded isn't equal to multipart/form-data; boundary="
+                in msg
+            )
+
+        # empty body request
+        with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
+            req_files = {"file_name": b"Old World!"}
+            rsps.add(
+                responses.POST,
+                url="http://httpbin.org/post",
+                match=[matchers.multipart_matcher(req_files)],
+            )
+
+            with pytest.raises(ConnectionError) as excinfo:
+                requests.post("http://httpbin.org/post")
+
+            msg = str(excinfo.value)
+            assert "Request is missing the 'Content-Type' header" in msg
+
+    run()
+    assert_reset()
+
+
+def test_query_string_matcher_raises():
+    """
+    Validate that Exception is raised if request does not match responses.matchers
+        validate matchers.query_string_matcher
+            :return: None
+    """
+
+    def run():
+        with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
+            rsps.add(
+                "GET",
+                "http://111.com",
+                match=[matchers.query_string_matcher("didi=pro")],
+            )
+
+            with pytest.raises(ConnectionError) as excinfo:
+                requests.get("http://111.com", params={"test": "1", "didi": "pro"})
+
+            msg = str(excinfo.value)
+            assert (
+                "Query string doesn't match. {didi: pro, test: 1} doesn't match {didi: pro}"
+                in msg
+            )
+
+    run()
+    assert_reset()
+
+
+def test_request_matches_headers():
+    @responses.activate
+    def run():
+        url = "http://example.com/"
         responses.add(
-            "GET",
-            "http://111.com",
-            match=[matchers.query_param_matcher({"my": "params"})],
+            method=responses.GET,
+            url=url,
+            json={"success": True},
+            match=[matchers.header_matcher({"Accept": "application/json"})],
         )
 
         responses.add(
             method=responses.GET,
-            url="http://111.com/",
-            body="two",
-            match=[matchers.json_params_matcher({"page": "one"})],
+            url=url,
+            body="success",
+            match=[matchers.header_matcher({"Accept": "text/plain"})],
+        )
+
+        # the actual request can contain extra headers (requests always adds some itself anyway)
+        resp = requests.get(
+            url, headers={"Accept": "application/json", "Accept-Charset": "utf-8"}
+        )
+        assert_response(resp, body='{"success": true}', content_type="application/json")
+
+        resp = requests.get(url, headers={"Accept": "text/plain"})
+        assert_response(resp, body="success", content_type="text/plain")
+
+    run()
+    assert_reset()
+
+
+def test_request_matches_headers_no_match():
+    @responses.activate
+    def run():
+        url = "http://example.com/"
+        responses.add(
+            method=responses.GET,
+            url=url,
+            json={"success": True},
+            match=[matchers.header_matcher({"Accept": "application/json"})],
         )
 
         with pytest.raises(ConnectionError) as excinfo:
-            requests.get("http://111.com", params={"id": "bad"}, json={"page": "two"})
+            requests.get(url, headers={"Accept": "application/xml"})
 
         msg = str(excinfo.value)
-        if six.PY3:
-            assert (
-                "Parameters do not match. [('id', 'bad')] doesn't match [('my', 'params')]"
-                in msg
+        assert (
+            "Headers do not match: {Accept: application/xml} doesn't match "
+            "{Accept: application/json}"
+        ) in msg
+
+    run()
+    assert_reset()
+
+
+def test_request_matches_headers_strict_match():
+    @responses.activate
+    def run():
+        url = "http://example.com/"
+        responses.add(
+            method=responses.GET,
+            url=url,
+            body="success",
+            match=[
+                matchers.header_matcher({"Accept": "text/plain"}, strict_match=True)
+            ],
+        )
+
+        # requests will add some extra headers of its own, so we have to use prepared requests
+        session = requests.Session()
+
+        # make sure we send *just* the header we're expectin
+        prepped = session.prepare_request(
+            requests.Request(
+                method="GET",
+                url=url,
             )
-            assert (
-                """Parameters do not match. {"page": "two"} doesn't match {'page': 'one'}"""
-                in msg
-            )
-        else:
-            assert (
-                "Parameters do not match. [('id', 'bad')] doesn't match [(u'my', u'params')]"
-                in msg
-            )
-            assert (
-                """Parameters do not match. {"page": "two"} doesn't match {u'page': u'one'}"""
-                in msg
+        )
+        prepped.headers.clear()
+        prepped.headers["Accept"] = "text/plain"
+
+        resp = session.send(prepped)
+        assert_response(resp, body="success", content_type="text/plain")
+
+        # include the "Accept-Charset" header, which will fail to match
+        prepped = session.prepare_request(
+            requests.Request(
+                method="GET",
+                url=url,
             )
+        )
+        prepped.headers.clear()
+        prepped.headers["Accept"] = "text/plain"
+        prepped.headers["Accept-Charset"] = "utf-8"
+
+        with pytest.raises(ConnectionError) as excinfo:
+            session.send(prepped)
+
+        msg = str(excinfo.value)
+        assert (
+            "Headers do not match: {Accept: text/plain, Accept-Charset: utf-8} "
+            "doesn't match {Accept: text/plain}"
+        ) in msg
 
     run()
     assert_reset()
@@ -1694,3 +2142,28 @@ def test_rfc_compliance(url, other_url):
 
     run()
     assert_reset()
+
+
+def test_matchers_create_key_val_str():
+    """
+    Test that matchers._create_key_val_str does recursive conversion
+    """
+    data = {
+        "my_list": [
+            1,
+            2,
+            "a",
+            {"key1": "val1", "key2": 2, 3: "test"},
+            "!",
+            [["list", "nested"], {"nested": "dict"}],
+        ],
+        1: 4,
+        "test": "val",
+        "high": {"nested": "nested_dict"},
+    }
+    conv_str = matchers._create_key_val_str(data)
+    reference = (
+        "{1: 4, high: {nested: nested_dict}, my_list: [!, 1, 2, [[list, nested], {nested: dict}], "
+        "a, {3: test, key1: val1, key2: 2}], test: val}"
+    )
+    assert conv_str == reference
